@@ -5,20 +5,19 @@
 #include <ctype.h>
 #include <errno.h>
 
+#include <sys/epoll.h>
+
 #define LOG_MODULE "osc"
 #define LOG_ENABLE_DBG 0
 #include "log.h"
 #include "base64.h"
 #include "config.h"
-#include "grid.h"
 #include "macros.h"
 #include "notify.h"
-#include "render.h"
 #include "selection.h"
 #include "terminal.h"
 #include "uri.h"
 #include "util.h"
-#include "vt.h"
 #include "xmalloc.h"
 #include "xsnprintf.h"
 
@@ -65,7 +64,7 @@ osc_to_clipboard(struct terminal *term, const char *target,
         return;
     }
 
-    char *decoded = base64_decode(base64_data);
+    char *decoded = base64_decode(base64_data, NULL);
     if (decoded == NULL) {
         if (errno == EINVAL)
             LOG_WARN("OSC: invalid clipboard data: %s", base64_data);
@@ -124,7 +123,7 @@ from_clipboard_cb(char *text, size_t size, void *user)
             xassert(chunk != NULL);
             xassert(strlen(chunk) == 4);
 
-            term_to_slave(term, chunk, 4);
+            term_paste_data_to_slave(term, chunk, 4);
             free(chunk);
 
             ctx->idx = 0;
@@ -144,7 +143,7 @@ from_clipboard_cb(char *text, size_t size, void *user)
     char *chunk = base64_encode((const uint8_t *)t, left / 3 * 3);
     xassert(chunk != NULL);
     xassert(strlen(chunk) % 4 == 0);
-    term_to_slave(term, chunk, strlen(chunk));
+    term_paste_data_to_slave(term, chunk, strlen(chunk));
     free(chunk);
 }
 
@@ -157,13 +156,19 @@ from_clipboard_done(void *user)
     if (ctx->idx > 0) {
         char res[4];
         base64_encode_final(ctx->buf, ctx->idx, res);
-        term_to_slave(term, res, 4);
+        term_paste_data_to_slave(term, res, 4);
     }
 
     if (term->vt.osc.bel)
-        term_to_slave(term, "\a", 1);
+        term_paste_data_to_slave(term, "\a", 1);
     else
-        term_to_slave(term, "\033\\", 2);
+        term_paste_data_to_slave(term, "\033\\", 2);
+
+    term->is_sending_paste_data = false;
+
+    /* Make sure we send any queued up non-paste data */
+    if (tll_length(term->ptmx_buffers) > 0)
+        fdm_event_add(term->fdm, term->ptmx, EPOLLOUT);
 
     free(ctx);
 }
@@ -214,9 +219,24 @@ osc_from_clipboard(struct terminal *term, const char *source)
     if (!from_clipboard && !from_primary)
         return;
 
-    term_to_slave(term, "\033]52;", 5);
-    term_to_slave(term, &src, 1);
-    term_to_slave(term, ";", 1);
+    if (term->is_sending_paste_data) {
+        /* FIXME: we should wait for the paste to end, then continue
+           with the OSC-52 reply */
+        term_to_slave(term, "\033]52;", 5);
+        term_to_slave(term, &src, 1);
+        term_to_slave(term, ";", 1);
+        if (term->vt.osc.bel)
+            term_to_slave(term, "\a", 1);
+        else
+            term_to_slave(term, "\033\\", 2);
+        return;
+    }
+
+    term->is_sending_paste_data = true;
+
+    term_paste_data_to_slave(term, "\033]52;", 5);
+    term_paste_data_to_slave(term, &src, 1);
+    term_paste_data_to_slave(term, ";", 1);
 
     struct clip_context *ctx = xmalloc(sizeof(*ctx));
     *ctx = (struct clip_context) {.seat = seat, .term = term};
@@ -524,7 +544,499 @@ osc_notify(struct terminal *term, char *string)
     const char *title = strtok_r(string, ";", &ctx);
     const char *msg = strtok_r(NULL, "\x00", &ctx);
 
-    notify_notify(term, title, msg != NULL ? msg : "");
+    if (title == NULL)
+        return;
+
+    if (mbsntoc32(NULL, title, strlen(title), 0) == (char32_t)-1) {
+        LOG_WARN("%s: notification title is not valid UTF-8, ignoring", title);
+        return;
+    }
+
+    if (msg != NULL && mbsntoc32(NULL, msg, strlen(msg), 0) == (char32_t)-1) {
+        LOG_WARN("%s: notification message is not valid UTF-8, ignoring", msg);
+        return;
+    }
+
+    notify_notify(term, &(struct notification){
+        .title = (char *)title,
+        .body = (char *)msg,
+        .expire_time = -1,
+    });
+}
+
+static void
+kitty_notification(struct terminal *term, char *string)
+{
+    /* https://sw.kovidgoyal.net/kitty/desktop-notifications */
+
+    char *payload_raw = strchr(string, ';');
+    if (payload_raw == NULL)
+        return;
+
+    char *parameters = string;
+    *payload_raw = '\0';
+    payload_raw++;
+
+    char *id = NULL;               /* The 'i' parameter */
+    char *app_id = NULL;           /* The 'f' parameter */
+    char *icon_cache_id = NULL;    /* The 'g' parameter */
+    char *symbolic_icon = NULL;    /* The 'n' parameter */
+    char *category = NULL;         /* The 't' parameter */
+    char *payload = NULL;
+
+    bool focus = true;             /* The 'a' parameter */
+    bool report_activated = false; /* The 'a' parameter */
+    bool report_closed = false;    /* The 'c' parameter */
+    bool done = true;              /* The 'd' parameter */
+    bool base64 = false;           /* The 'e' parameter */
+
+    int32_t expire_time = -1;      /* The 'w' parameter */
+
+    size_t payload_size;
+    enum {
+        PAYLOAD_TITLE,
+        PAYLOAD_BODY,
+        PAYLOAD_CLOSE,
+        PAYLOAD_ALIVE,
+        PAYLOAD_ICON,
+        PAYLOAD_BUTTON,
+    } payload_type = PAYLOAD_TITLE; /* The 'p' parameter */
+
+    enum notify_when when = NOTIFY_ALWAYS;
+    enum notify_urgency urgency = NOTIFY_URGENCY_NORMAL;
+
+    bool have_a = false;
+    bool have_c = false;
+    bool have_o = false;
+    bool have_u = false;
+    bool have_w = false;
+
+    char *ctx = NULL;
+    for (char *param = strtok_r(parameters, ":", &ctx);
+         param != NULL;
+         param = strtok_r(NULL, ":", &ctx))
+    {
+        /* All parameters are on the form X=value, where X is always
+           exactly one character */
+        if (param[0] == '\0' || param[1] != '=')
+            continue;
+
+        char *value = &param[2];
+
+        switch (param[0]) {
+        case 'a': {
+            /* notification activation action: focus|report|-focus|-report */
+            have_a = true;
+            char *a_ctx = NULL;
+
+            for (const char *v = strtok_r(value, ",", &a_ctx);
+                 v != NULL;
+                 v = strtok_r(NULL, ",", &a_ctx))
+            {
+                bool reverse = v[0] == '-';
+                if (reverse)
+                    v++;
+
+                if (streq(v, "focus"))
+                    focus = !reverse;
+                else if (streq(v, "report"))
+                    report_activated = !reverse;
+            }
+
+            break;
+        }
+
+        case 'c':
+            if (value[0] == '1' && value[1] == '\0')
+                report_closed = true;
+            else if (value[0] == '0' && value[1] == '\0')
+                report_closed = false;
+            have_c = true;
+            break;
+
+        case 'd':
+            /* done: 0|1 */
+            if (value[0] == '0' && value[1] == '\0')
+                done = false;
+            else if (value[0] == '1' && value[1] == '\0')
+                done = true;
+            break;
+
+        case 'e':
+            /* base64 (payload encoding): 0=utf8, 1=base64(utf8) */
+            if (value[0] == '0' && value[1] == '\0')
+                base64 = false;
+            else if (value[0] == '1' && value[1] == '\0')
+                base64 = true;
+            break;
+
+        case 'i':
+            /* id */
+            free(id);
+            id = xstrdup(value);
+            break;
+
+        case 'p':
+            /* payload content: title|body */
+            if (streq(value, "title"))
+                payload_type = PAYLOAD_TITLE;
+            else if (streq(value, "body"))
+                payload_type = PAYLOAD_BODY;
+            else if (streq(value, "close"))
+                payload_type = PAYLOAD_CLOSE;
+            else if (streq(value, "alive"))
+                payload_type = PAYLOAD_ALIVE;
+            else if (streq(value, "icon"))
+                payload_type = PAYLOAD_ICON;
+            else if (streq(value, "buttons"))
+                payload_type = PAYLOAD_BUTTON;
+            else if (streq(value, "?")) {
+                /* Query capabilities */
+
+                const char *reply_id = id != NULL ? id : "0";
+
+                const char *p_caps = "title,body,?,close,alive,icon,buttons";
+                const char *a_caps = "focus,report";
+                const char *u_caps = "0,1,2";
+
+                char when_caps[64];
+                strcpy(when_caps, "unfocused");
+                if (!term->conf->desktop_notifications.inhibit_when_focused)
+                    strcat(when_caps, ",always");
+
+                const char *terminator = term->vt.osc.bel ? "\a" : "\033\\";
+
+                char reply[128];
+                int n = xsnprintf(
+                    reply, sizeof(reply),
+                    "\033]99;i=%s:p=?;p=%s:a=%s:o=%s:u=%s:c=1:w=1%s",
+                    reply_id, p_caps, a_caps, when_caps, u_caps, terminator);
+
+                xassert(n < sizeof(reply));
+                term_to_slave(term, reply, n);
+                goto out;
+            }
+            break;
+
+        case 'o':
+            /* honor when: always|unfocused|invisible */
+            have_o = true;
+            if (streq(value, "always"))
+                when = NOTIFY_ALWAYS;
+            else if (streq(value, "unfocused"))
+                when = NOTIFY_UNFOCUSED;
+            else if (streq(value, "invisible"))
+                when = NOTIFY_INVISIBLE;
+            break;
+
+        case 'u':
+            /* urgency: 0=low, 1=normal, 2=critical */
+            have_u = true;
+            if (value[0] == '0' && value[1] == '\0')
+                urgency = NOTIFY_URGENCY_LOW;
+            else if (value[0] == '1' && value[1] == '\0')
+                urgency = NOTIFY_URGENCY_NORMAL;
+            else if (value[0] == '2' && value[1] == '\0')
+                urgency = NOTIFY_URGENCY_CRITICAL;
+            break;
+
+        case 'w': {
+            /* Notification timeout */
+            errno = 0;
+            char *end = NULL;
+            long timeout = strtol(value, &end, 10);
+
+            if (errno == 0 && *end == '\0' && timeout <= INT32_MAX) {
+                expire_time = timeout;
+                have_w = true;
+            }
+            break;
+        }
+
+        case 'f':
+            free(app_id);
+            app_id = base64_decode(value, NULL);
+            break;
+
+        case 't': {
+            /* Type (category) */
+            char *decoded = base64_decode(value, NULL);
+            if (decoded != NULL) {
+                if (category == NULL)
+                    category = decoded;
+                else {
+                    /* Append, comma separated */
+                    char *old_category = category;
+                    category = xstrjoin(old_category, decoded, ',');
+                    free(decoded);
+                    free(old_category);
+                }
+            }
+            break;
+        }
+
+        case 'g':
+            /* graphical ID (see 'n' and 'p=icon') */
+            free(icon_cache_id);
+            icon_cache_id = xstrdup(value);
+            break;
+
+        case 'n': {
+            /* Symbolic icon name, may used with 'g' */
+
+            /*
+             * Sigh, protocol says 'n' can be used multiple times, and
+             * that the terminal picks the first one that it can
+             * resolve.
+             *
+             * We can't resolve any icons at all. So, enter
+             * heuristics... let's pick the *shortest* symbolic
+             * name. The idea is that icon *names* are typically
+             * shorter than .desktop names, and macOS bundle
+             * identifiers.
+             */
+            char *maybe_new_symbolic_icon = base64_decode(value, NULL);
+            if (maybe_new_symbolic_icon == NULL)
+                break;
+
+            if (symbolic_icon == NULL ||
+                strlen(maybe_new_symbolic_icon) < strlen(symbolic_icon))
+            {
+                free(symbolic_icon);
+                symbolic_icon = maybe_new_symbolic_icon;
+
+                /* Translate OSC-99 "special" names */
+                if (symbolic_icon != NULL) {
+                    const char *translated_name = NULL;
+
+                    if (streq(symbolic_icon, "error"))
+                        translated_name = "dialog-error";
+                    else if (streq(symbolic_icon, "warn") ||
+                             streq(symbolic_icon, "warning"))
+                        translated_name = "dialog-warning";
+                    else if (streq(symbolic_icon, "info"))
+                        translated_name = "dialog-information";
+                    else if (streq(symbolic_icon, "question"))
+                        translated_name = "dialog-question";
+                    else if (streq(symbolic_icon, "help"))
+                        translated_name = "system-help";
+                    else if (streq(symbolic_icon, "file-manager"))
+                        translated_name = "system-file-manager";
+                    else if (streq(symbolic_icon, "system-monitor"))
+                        translated_name = "utilities-system-monitor";
+                    else if (streq(symbolic_icon, "text-editor"))
+                        translated_name = "text-editor";
+
+                    if (translated_name != NULL) {
+                        free(symbolic_icon);
+                        symbolic_icon = xstrdup(translated_name);
+                    }
+                }
+            } else {
+                free(maybe_new_symbolic_icon);
+            }
+            break;
+        }
+        }
+    }
+
+    if (base64) {
+        payload = base64_decode(payload_raw, &payload_size);
+        if (payload == NULL)
+            goto out;
+    } else {
+        payload = xstrdup(payload_raw);
+        payload_size = strlen(payload);
+    }
+
+    /* Append metadata to previous notification chunk */
+    struct notification *notif = &term->kitty_notification;
+
+    if (!((id == NULL && notif->id == NULL) ||
+          (id != NULL && notif->id != NULL && streq(id, notif->id))) ||
+        !notif->may_be_programatically_closed)  /* Free:d notification has this as false... */
+    {
+        /* ID mismatch, ignore previous notification state */
+        notify_free(term, notif);
+
+        notif->id = id;
+        notif->when = when;
+        notif->urgency = urgency;
+        notif->expire_time = expire_time;
+        notif->focus = focus;
+        notif->may_be_programatically_closed = true;
+        notif->report_activated = report_activated;
+        notif->report_closed = report_closed;
+
+        id = NULL; /* Prevent double free */
+    }
+
+    if (have_a) {
+        notif->focus = focus;
+        notif->report_activated = report_activated;
+    }
+
+    if (have_c)
+        notif->report_closed = report_closed;
+
+    if (have_o)
+        notif->when = when;
+    if (have_u)
+        notif->urgency = urgency;
+    if (have_w)
+        notif->expire_time = expire_time;
+
+    if (icon_cache_id != NULL) {
+        free(notif->icon_cache_id);
+        notif->icon_cache_id = icon_cache_id;
+        icon_cache_id = NULL;  /* Prevent double free */
+    }
+
+    if (symbolic_icon != NULL) {
+        free(notif->icon_symbolic_name);
+        notif->icon_symbolic_name = symbolic_icon;
+        symbolic_icon = NULL;
+    }
+
+    if (app_id != NULL) {
+        free(notif->app_id);
+        notif->app_id = app_id;
+        app_id = NULL;  /* Prevent double free */
+    }
+
+    if (category != NULL) {
+        if (notif->category == NULL) {
+            notif->category = category;
+            category = NULL;  /* Prevent double free */
+        } else {
+            /* Append, comma separated */
+            char *new_category = xstrjoin(notif->category, category, ',');
+            free(notif->category);
+            notif->category = new_category;
+        }
+    }
+
+    /* Handled chunked payload - append to existing metadata */
+    switch (payload_type) {
+    case PAYLOAD_TITLE:
+    case PAYLOAD_BODY: {
+        char **ptr = payload_type == PAYLOAD_TITLE
+            ? &notif->title
+            : &notif->body;
+
+        if (*ptr == NULL) {
+            *ptr = payload;
+            payload = NULL;
+        } else {
+            char *old = *ptr;
+            *ptr = xstrjoin(old, payload, 0);
+            free(old);
+        }
+        break;
+    }
+
+    case PAYLOAD_CLOSE:
+    case PAYLOAD_ALIVE:
+        /* Ignore payload */
+        break;
+
+    case PAYLOAD_ICON:
+        if (notif->icon_data == NULL) {
+            notif->icon_data = (uint8_t *)payload;
+            notif->icon_data_sz = payload_size;
+            payload = NULL;
+        } else {
+            notif->icon_data = xrealloc(
+                notif->icon_data, notif->icon_data_sz + payload_size);
+            memcpy(&notif->icon_data[notif->icon_data_sz], payload, payload_size);
+            notif->icon_data_sz += payload_size;
+        }
+        break;
+
+    case PAYLOAD_BUTTON: {
+        char *ctx = NULL;
+        for (const char *button = strtok_r(payload, "\u2028", &ctx);
+             button != NULL;
+             button = strtok_r(NULL, "\u2028", &ctx))
+        {
+            if (button[0] != '\0') {
+                tll_push_back(notif->actions, xstrdup(button));
+            }
+        }
+        
+        break;
+    }
+    }
+
+    if (done) {
+        /* Update icon cache, if necessary */
+        if (notif->icon_cache_id != NULL &&
+            (notif->icon_symbolic_name != NULL || notif->icon_data != NULL))
+        {
+            notify_icon_del(term, notif->icon_cache_id);
+            notify_icon_add(term, notif->icon_cache_id,
+                            notif->icon_symbolic_name,
+                            notif->icon_data, notif->icon_data_sz);
+
+            /* Don't need this anymore */
+            free(notif->icon_symbolic_name);
+            free(notif->icon_data);
+            notif->icon_symbolic_name = NULL;
+            notif->icon_data = NULL;
+            notif->icon_data_sz = 0;
+        }
+
+        if (payload_type == PAYLOAD_CLOSE) {
+            if (notif->id != NULL)
+                notify_close(term, notif->id);
+        } else if (payload_type == PAYLOAD_ALIVE)  {
+            char *alive_ids = NULL;
+
+            tll_foreach(term->active_notifications, it) {
+                /* TODO: check with kitty: use "0" for all
+                   notifications with no ID? */
+
+                const char *item_id = it->item.id != NULL ? it->item.id : "0";
+
+                if (alive_ids == NULL)
+                    alive_ids = xstrdup(item_id);
+                else {
+                    char *old_alive_ids = alive_ids;
+                    alive_ids = xstrjoin(old_alive_ids, item_id, ',');
+                    free(old_alive_ids);
+                }
+            }
+
+            char *reply = xasprintf(
+                "\033]99;i=%s:p=alive;%s\033\\",
+                notif->id != NULL ? notif->id : "0",
+                alive_ids != NULL ? alive_ids : "");
+
+            term_to_slave(term, reply, strlen(reply));
+            free(reply);
+            free(alive_ids);
+        } else {
+            /*
+             * Show notification.
+             *
+             * The checks for title|body is to handle notifications that
+             * only load icon data into the icon cache
+             */
+            if (notif->title != NULL || notif->body != NULL) {
+                notify_notify(term, notif);
+            }
+        }
+
+        notify_free(term, notif);
+    }
+
+out:
+    free(id);
+    free(app_id);
+    free(icon_cache_id);
+    free(symbolic_icon);
+    free(payload);
+    free(category);
 }
 
 void
@@ -615,47 +1127,7 @@ osc_dispatch(struct terminal *term)
                         idx, term->colors.table[idx], color);
 
                 term->colors.table[idx] = color;
-
-                /* Dirty visible, affected cells */
-                for (int r = 0; r < term->rows; r++) {
-                    struct row *row = grid_row_in_view(term->grid, r);
-                    struct cell *cell = &row->cells[0];
-
-                    for (int c = 0; c < term->cols; c++, cell++) {
-                        bool dirty = false;
-
-                        switch (cell->attrs.fg_src) {
-                        case COLOR_BASE16:
-                        case COLOR_BASE256:
-                            if (cell->attrs.fg == idx)
-                                dirty = true;
-                            break;
-
-                        case COLOR_DEFAULT:
-                        case COLOR_RGB:
-                            /* Not affected */
-                            break;
-                        }
-
-                        switch (cell->attrs.bg_src) {
-                        case COLOR_BASE16:
-                        case COLOR_BASE256:
-                            if (cell->attrs.bg == idx)
-                                dirty = true;
-                            break;
-
-                        case COLOR_DEFAULT:
-                        case COLOR_RGB:
-                            /* Not affected */
-                            break;
-                        }
-
-                        if (dirty) {
-                            cell->attrs.clean = 0;
-                            row->dirty = true;
-                        }
-                    }
-                }
+                term_damage_color(term, COLOR_BASE256, idx);
             }
         }
 
@@ -676,15 +1148,25 @@ osc_dispatch(struct terminal *term)
         osc_notify(term, string);
         break;
 
-    case 10:
-    case 11:
-    case 17:
-    case 19: {
+    case 10:    /* fg */
+    case 11:    /* bg */
+    case 12:    /* cursor */
+    case 17:    /* highlight (selection) fg */
+    case 19: {  /* highlight (selection) bg */
         /* Set default foreground/background/highlight-bg/highlight-fg color */
 
         /* Client queried for current value */
         if (string[0] == '?' && string[1] == '\0') {
-            uint32_t color = param == 10 ? term->colors.fg : term->colors.bg;
+            uint32_t color = param == 10
+                ? term->colors.fg
+                : param == 11
+                    ? term->colors.bg
+                    : param == 12
+                        ? term->colors.cursor_bg
+                        : param == 17
+                            ? term->colors.selection_bg
+                            : term->colors.selection_fg;
+
             uint8_t r = (color >> 16) & 0xff;
             uint8_t g = (color >>  8) & 0xff;
             uint8_t b = (color >>  0) & 0xff;
@@ -718,6 +1200,7 @@ osc_dispatch(struct terminal *term)
         LOG_DBG("change color definition for %s to %06x",
                 param == 10 ? "foreground" :
                 param == 11 ? "background" :
+                param == 12 ? "cursor" :
                 param == 17 ? "selection background" :
                               "selection foreground",
                 color);
@@ -725,6 +1208,7 @@ osc_dispatch(struct terminal *term)
         switch (param) {
         case 10:
             term->colors.fg = color;
+            term_damage_color(term, COLOR_DEFAULT, 0);
             break;
 
         case 11:
@@ -738,6 +1222,13 @@ osc_dispatch(struct terminal *term)
                     term_font_subpixel_changed(term);
                 }
             }
+            term_damage_color(term, COLOR_DEFAULT, 0);
+            term_damage_margins(term);
+            break;
+
+        case 12:
+            term->colors.cursor_bg = 1u << 31 | color;
+            term_damage_cursor(term);
             break;
 
         case 17:
@@ -751,48 +1242,9 @@ osc_dispatch(struct terminal *term)
             break;
         }
 
-        term_damage_view(term);
-        term_damage_margins(term);
         break;
     }
 
-    case 12: /* Set cursor color */
-
-        /* Client queried for current value */
-        if (string[0] == '?' && string[1] == '\0') {
-            uint8_t r = (term->cursor_color.cursor >> 16) & 0xff;
-            uint8_t g = (term->cursor_color.cursor >>  8) & 0xff;
-            uint8_t b = (term->cursor_color.cursor >>  0) & 0xff;
-            const char *terminator = term->vt.osc.bel ? "\a" : "\033\\";
-
-            char reply[32];
-            size_t n = xsnprintf(
-                reply, sizeof(reply), "\033]12;rgb:%02x/%02x/%02x%s",
-                r, g, b, terminator);
-
-            term_to_slave(term, reply, n);
-            break;
-        }
-
-        uint32_t color;
-
-        if (string[0] == '#' || string[0] == '['
-            ? !parse_legacy_color(string, &color, NULL, NULL)
-            : !parse_rgb(string, &color, NULL, NULL))
-        {
-            break;
-        }
-
-        LOG_DBG("change cursor color to %06x", color);
-
-        if (color == 0)
-            term->cursor_color.cursor = 0;  /* Invert fg/bg */
-        else
-            term->cursor_color.cursor = 1u << 31 | color;
-
-        term_damage_cursor(term);
-        break;
-
     case 22:  /* Set mouse cursor */
         term_set_user_mouse_cursor(term, string);
         break;
@@ -804,6 +1256,10 @@ osc_dispatch(struct terminal *term)
         osc_selection(term, string);
         break;
 
+    case 99:  /* Kitty notifications */
+        kitty_notification(term, string);
+        break;
+
     case 104: {
         /* Reset Color Number 'c' (whole table if no parameter) */
 
@@ -811,6 +1267,7 @@ osc_dispatch(struct terminal *term)
             LOG_DBG("resetting all colors");
             for (size_t i = 0; i < ALEN(term->colors.table); i++)
                 term->colors.table[i] = term->conf->colors.table[i];
+            term_damage_view(term);
         }
 
         else {
@@ -832,11 +1289,10 @@ osc_dispatch(struct terminal *term)
 
                 LOG_DBG("resetting color #%u", idx);
                 term->colors.table[idx] = term->conf->colors.table[idx];
+                term_damage_color(term, COLOR_BASE256, idx);
             }
 
         }
-
-        term_damage_view(term);
         break;
     }
 
@@ -846,21 +1302,21 @@ osc_dispatch(struct terminal *term)
     case 110: /* Reset default text foreground color */
         LOG_DBG("resetting foreground color");
         term->colors.fg = term->conf->colors.fg;
-        term_damage_view(term);
+        term_damage_color(term, COLOR_DEFAULT, 0);
         break;
 
     case 111: /* Reset default text background color */
         LOG_DBG("resetting background color");
         term->colors.bg = term->conf->colors.bg;
         term->colors.alpha = term->conf->colors.alpha;
-        term_damage_view(term);
+        term_damage_color(term, COLOR_DEFAULT, 0);
         term_damage_margins(term);
         break;
 
     case 112:
         LOG_DBG("resetting cursor color");
-        term->cursor_color.text = term->conf->cursor.color.text;
-        term->cursor_color.cursor = term->conf->cursor.color.cursor;
+        term->colors.cursor_fg = term->conf->cursor.color.text;
+        term->colors.cursor_bg = term->conf->cursor.color.cursor;
         term_damage_cursor(term);
         break;
 
